@@ -47,9 +47,6 @@
 
 #define max(a,b) (((a) > (b)) ? (a) : (b))
 
-#define COMMAND_LENGTH 2048
-#define WORKGROUP_STRING_LENGTH 1024
-
 struct data {
   /* Currently loaded kernel. */
   cl_kernel current_kernel;
@@ -64,103 +61,21 @@ struct data {
 };
 
 static const cl_image_format supported_image_formats[] = {
-    { CL_R, CL_SNORM_INT8 },
-    { CL_R, CL_SNORM_INT16 },
-    { CL_R, CL_UNORM_INT8 },
-    { CL_R, CL_UNORM_INT16 },
-    { CL_R, CL_UNORM_SHORT_565 }, 
-    { CL_R, CL_UNORM_SHORT_555 },
-    { CL_R, CL_UNORM_INT_101010 }, 
-    { CL_R, CL_SIGNED_INT8 },
-    { CL_R, CL_SIGNED_INT16 }, 
-    { CL_R, CL_SIGNED_INT32 },
-    { CL_R, CL_UNSIGNED_INT8 }, 
-    { CL_R, CL_UNSIGNED_INT16 },
-    { CL_R, CL_UNSIGNED_INT32 }, 
-    { CL_R, CL_HALF_FLOAT },
-    { CL_R, CL_FLOAT },
-    { CL_Rx, CL_SNORM_INT8 },
-    { CL_Rx, CL_SNORM_INT16 },
-    { CL_Rx, CL_UNORM_INT8 },
-    { CL_Rx, CL_UNORM_INT16 },
-    { CL_Rx, CL_UNORM_SHORT_565 }, 
-    { CL_Rx, CL_UNORM_SHORT_555 },
-    { CL_Rx, CL_UNORM_INT_101010 }, 
-    { CL_Rx, CL_SIGNED_INT8 },
-    { CL_Rx, CL_SIGNED_INT16 }, 
-    { CL_Rx, CL_SIGNED_INT32 },
-    { CL_Rx, CL_UNSIGNED_INT8 }, 
-    { CL_Rx, CL_UNSIGNED_INT16 },
-    { CL_Rx, CL_UNSIGNED_INT32 }, 
-    { CL_Rx, CL_HALF_FLOAT },
-    { CL_Rx, CL_FLOAT },
     { CL_A, CL_SNORM_INT8 },
     { CL_A, CL_SNORM_INT16 },
     { CL_A, CL_UNORM_INT8 },
     { CL_A, CL_UNORM_INT16 },
-    { CL_A, CL_UNORM_SHORT_565 }, 
-    { CL_A, CL_UNORM_SHORT_555 },
-    { CL_A, CL_UNORM_INT_101010 }, 
     { CL_A, CL_SIGNED_INT8 },
     { CL_A, CL_SIGNED_INT16 }, 
     { CL_A, CL_SIGNED_INT32 },
     { CL_A, CL_UNSIGNED_INT8 }, 
     { CL_A, CL_UNSIGNED_INT16 },
     { CL_A, CL_UNSIGNED_INT32 }, 
-    { CL_A, CL_HALF_FLOAT },
     { CL_A, CL_FLOAT },
-    { CL_RG, CL_SNORM_INT8 },
-    { CL_RG, CL_SNORM_INT16 },
-    { CL_RG, CL_UNORM_INT8 },
-    { CL_RG, CL_UNORM_INT16 },
-    { CL_RG, CL_UNORM_SHORT_565 }, 
-    { CL_RG, CL_UNORM_SHORT_555 },
-    { CL_RG, CL_UNORM_INT_101010 }, 
-    { CL_RG, CL_SIGNED_INT8 },
-    { CL_RG, CL_SIGNED_INT16 }, 
-    { CL_RG, CL_SIGNED_INT32 },
-    { CL_RG, CL_UNSIGNED_INT8 }, 
-    { CL_RG, CL_UNSIGNED_INT16 },
-    { CL_RG, CL_UNSIGNED_INT32 }, 
-    { CL_RG, CL_HALF_FLOAT },
-    { CL_RG, CL_FLOAT },
-    { CL_RGx, CL_SNORM_INT8 },
-    { CL_RGx, CL_SNORM_INT16 },
-    { CL_RGx, CL_UNORM_INT8 },
-    { CL_RGx, CL_UNORM_INT16 },
-    { CL_RGx, CL_UNORM_SHORT_565 }, 
-    { CL_RGx, CL_UNORM_SHORT_555 },
-    { CL_RGx, CL_UNORM_INT_101010 }, 
-    { CL_RGx, CL_SIGNED_INT8 },
-    { CL_RGx, CL_SIGNED_INT16 }, 
-    { CL_RGx, CL_SIGNED_INT32 },
-    { CL_RGx, CL_UNSIGNED_INT8 }, 
-    { CL_RGx, CL_UNSIGNED_INT16 },
-    { CL_RGx, CL_UNSIGNED_INT32 }, 
-    { CL_RGx, CL_HALF_FLOAT },
-    { CL_RGx, CL_FLOAT },
-    { CL_RA, CL_SNORM_INT8 },
-    { CL_RA, CL_SNORM_INT16 },
-    { CL_RA, CL_UNORM_INT8 },
-    { CL_RA, CL_UNORM_INT16 },
-    { CL_RA, CL_UNORM_SHORT_565 }, 
-    { CL_RA, CL_UNORM_SHORT_555 },
-    { CL_RA, CL_UNORM_INT_101010 }, 
-    { CL_RA, CL_SIGNED_INT8 },
-    { CL_RA, CL_SIGNED_INT16 }, 
-    { CL_RA, CL_SIGNED_INT32 },
-    { CL_RA, CL_UNSIGNED_INT8 }, 
-    { CL_RA, CL_UNSIGNED_INT16 },
-    { CL_RA, CL_UNSIGNED_INT32 }, 
-    { CL_RA, CL_HALF_FLOAT },
-    { CL_RA, CL_FLOAT },
     { CL_RGBA, CL_SNORM_INT8 },
     { CL_RGBA, CL_SNORM_INT16 },
     { CL_RGBA, CL_UNORM_INT8 },
     { CL_RGBA, CL_UNORM_INT16 },
-    { CL_RGBA, CL_UNORM_SHORT_565 }, 
-    { CL_RGBA, CL_UNORM_SHORT_555 },
-    { CL_RGBA, CL_UNORM_INT_101010 }, 
     { CL_RGBA, CL_SIGNED_INT8 },
     { CL_RGBA, CL_SIGNED_INT16 }, 
     { CL_RGBA, CL_SIGNED_INT32 },
@@ -169,24 +84,6 @@ static const cl_image_format supported_image_formats[] = {
     { CL_RGBA, CL_UNSIGNED_INT32 }, 
     { CL_RGBA, CL_HALF_FLOAT },
     { CL_RGBA, CL_FLOAT },
-    { CL_INTENSITY, CL_UNORM_INT8 }, 
-    { CL_INTENSITY, CL_UNORM_INT16 }, 
-    { CL_INTENSITY, CL_SNORM_INT8 }, 
-    { CL_INTENSITY, CL_SNORM_INT16 }, 
-    { CL_INTENSITY, CL_HALF_FLOAT }, 
-    { CL_INTENSITY, CL_FLOAT },
-    { CL_LUMINANCE, CL_UNORM_INT8 }, 
-    { CL_LUMINANCE, CL_UNORM_INT16 }, 
-    { CL_LUMINANCE, CL_SNORM_INT8 }, 
-    { CL_LUMINANCE, CL_SNORM_INT16 }, 
-    { CL_LUMINANCE, CL_HALF_FLOAT }, 
-    { CL_LUMINANCE, CL_FLOAT },
-    { CL_RGB, CL_UNORM_SHORT_565 }, 
-    { CL_RGB, CL_UNORM_SHORT_555 },
-    { CL_RGB, CL_UNORM_INT_101010 }, 
-    { CL_RGBx, CL_UNORM_SHORT_565 }, 
-    { CL_RGBx, CL_UNORM_SHORT_555 },
-    { CL_RGBx, CL_UNORM_INT_101010 }, 
     { CL_ARGB, CL_SNORM_INT8 },
     { CL_ARGB, CL_UNORM_INT8 },
     { CL_ARGB, CL_SIGNED_INT8 },
@@ -241,8 +138,10 @@ pocl_basic_build_hash (cl_device_id device)
   return res;
 }
 
+static cl_device_partition_property basic_partition_properties[1] = { 0 };
+
 void
-pocl_basic_init_device_infos(struct _cl_device_id* dev)
+pocl_basic_init_device_infos(unsigned j, struct _cl_device_id* dev)
 {
   dev->type = CL_DEVICE_TYPE_CPU;
   dev->vendor_id = 0;
@@ -257,10 +156,13 @@ pocl_basic_init_device_infos(struct _cl_device_id* dev)
     the SIMD lanes times the vector units, but not more than
     that to avoid stack overflow and cache trashing.
   */
-  dev->max_work_item_sizes[0] = dev->max_work_item_sizes[1] =
-	  dev->max_work_item_sizes[2] = dev->max_work_group_size = 1024*4;
+  dev->max_work_item_sizes[0] = dev->max_work_item_sizes[1]
+      = dev->max_work_item_sizes[2] = dev->max_work_group_size = 1024 * 4;
 
   dev->preferred_wg_size_multiple = 8;
+#ifdef OCS_AVAILABLE
+  cpu_setup_vector_widths (dev);
+#else
   dev->preferred_vector_width_char = POCL_DEVICES_PREFERRED_VECTOR_WIDTH_CHAR;
   dev->preferred_vector_width_short = POCL_DEVICES_PREFERRED_VECTOR_WIDTH_SHORT;
   dev->preferred_vector_width_int = POCL_DEVICES_PREFERRED_VECTOR_WIDTH_INT;
@@ -269,13 +171,15 @@ pocl_basic_init_device_infos(struct _cl_device_id* dev)
   dev->preferred_vector_width_double = POCL_DEVICES_PREFERRED_VECTOR_WIDTH_DOUBLE;
   dev->preferred_vector_width_half = POCL_DEVICES_PREFERRED_VECTOR_WIDTH_HALF;
   /* TODO: figure out what the difference between preferred and native widths are */
-  dev->native_vector_width_char = POCL_DEVICES_PREFERRED_VECTOR_WIDTH_CHAR;
-  dev->native_vector_width_short = POCL_DEVICES_PREFERRED_VECTOR_WIDTH_SHORT;
-  dev->native_vector_width_int = POCL_DEVICES_PREFERRED_VECTOR_WIDTH_INT;
-  dev->native_vector_width_long = POCL_DEVICES_PREFERRED_VECTOR_WIDTH_LONG;
-  dev->native_vector_width_float = POCL_DEVICES_PREFERRED_VECTOR_WIDTH_FLOAT;
-  dev->native_vector_width_double = POCL_DEVICES_PREFERRED_VECTOR_WIDTH_DOUBLE;
-  dev->native_vector_width_half = POCL_DEVICES_PREFERRED_VECTOR_WIDTH_HALF;
+  dev->native_vector_width_char = POCL_DEVICES_NATIVE_VECTOR_WIDTH_CHAR;
+  dev->native_vector_width_short = POCL_DEVICES_NATIVE_VECTOR_WIDTH_SHORT;
+  dev->native_vector_width_int = POCL_DEVICES_NATIVE_VECTOR_WIDTH_INT;
+  dev->native_vector_width_long = POCL_DEVICES_NATIVE_VECTOR_WIDTH_LONG;
+  dev->native_vector_width_float = POCL_DEVICES_NATIVE_VECTOR_WIDTH_FLOAT;
+  dev->native_vector_width_double = POCL_DEVICES_NATIVE_VECTOR_WIDTH_DOUBLE;
+  dev->native_vector_width_half = POCL_DEVICES_NATIVE_VECTOR_WIDTH_HALF;
+#endif
+
   dev->max_clock_frequency = 0;
   dev->address_bits = POCL_DEVICE_ADDRESS_BITS;
   dev->image_support = CL_TRUE;
@@ -295,7 +199,29 @@ pocl_basic_init_device_infos(struct _cl_device_id* dev)
   dev->mem_base_addr_align = MAX_EXTENDED_ALIGNMENT*8; // this is in bits
   dev->half_fp_config = 0;
   dev->single_fp_config = CL_FP_ROUND_TO_NEAREST | CL_FP_INF_NAN;
-  dev->double_fp_config = CL_FP_ROUND_TO_NEAREST | CL_FP_INF_NAN;
+#ifdef __x86_64__
+  dev->single_fp_config |= (CL_FP_DENORM | CL_FP_ROUND_TO_INF
+                            | CL_FP_ROUND_TO_ZERO
+                            | CL_FP_CORRECTLY_ROUNDED_DIVIDE_SQRT);
+#ifdef OCS_AVAILABLE
+  if (cpu_has_fma())
+    dev->single_fp_config |= CL_FP_FMA;
+#endif
+#endif
+
+#ifdef _CL_DISABLE_DOUBLE
+  dev->double_fp_config = 0;
+#else
+  /* TODO: all of these are the minimum mandated, but not all CPUs may actually
+   * support all of them. */
+  dev->double_fp_config = CL_FP_FMA | CL_FP_ROUND_TO_NEAREST
+                          | CL_FP_ROUND_TO_ZERO | CL_FP_ROUND_TO_INF
+                          | CL_FP_INF_NAN | CL_FP_DENORM;
+  /* this is a workaround for issue 28 in https://github.com/Oblomov/clinfo
+   * https://github.com/Oblomov/clinfo/issues/28 */
+  dev->double_fp_config |= CL_FP_CORRECTLY_ROUNDED_DIVIDE_SQRT;
+#endif
+
   dev->global_mem_cache_type = CL_NONE;
   dev->global_mem_cacheline_size = 0;
   dev->global_mem_cache_size = 0;
@@ -313,15 +239,23 @@ pocl_basic_init_device_infos(struct _cl_device_id* dev)
   dev->available = CL_TRUE;
   dev->compiler_available = CL_TRUE;
   dev->spmd = CL_FALSE;
+  dev->workgroup_pass = CL_TRUE;
   dev->execution_capabilities = CL_EXEC_KERNEL | CL_EXEC_NATIVE_KERNEL;
   dev->platform = 0;
 
   dev->parent_device = NULL;
-  // basic does not support partitioning
+  /* These two are only used for subdevices.
+   * Each subdevice has these two setup when created.
+   * The subdevice will then use these CUs:
+   *  [start, start+1, ..., start+count-1]
+   * this may not work with more complicated partitioning schemes,
+   * but is good enough for now. */
+  dev->core_start = 0;
+  dev->core_count = 0;
+  /* basic does not support partitioning */
   dev->max_sub_devices = 1;
   dev->num_partition_properties = 1;
-  dev->partition_properties = calloc(dev->num_partition_properties,
-    sizeof(cl_device_partition_property));
+  dev->partition_properties = basic_partition_properties;
   dev->num_partition_types = 0;
   dev->partition_type = NULL;
 
@@ -362,12 +296,7 @@ pocl_basic_init_device_infos(struct _cl_device_id* dev)
 #ifdef OCS_AVAILABLE
 
   dev->llvm_target_triplet = OCL_KERNEL_TARGET;
-
-#ifdef POCL_BUILT_WITH_CMAKE
   dev->llvm_cpu = get_cpu_name();
-#else
-  dev->llvm_cpu = OCL_KERNEL_TARGET_CPU;
-#endif
 
 #else
   dev->llvm_cpu = NULL;
@@ -392,13 +321,14 @@ pocl_basic_probe(struct pocl_device_ops *ops)
 
 
 
-void
-pocl_basic_init (cl_device_id device, const char* parameters)
+cl_int
+pocl_basic_init (unsigned j, cl_device_id device, const char* parameters)
 {
   struct data *d;
+  cl_int ret = CL_SUCCESS;
+  int err;
   static int first_basic_init = 1;
-  static int device_number = 0;
-  
+
   if (first_basic_init)
     {
       pocl_init_dlhandle_cache();
@@ -407,7 +337,9 @@ pocl_basic_init (cl_device_id device, const char* parameters)
   device->global_mem_id = 0;
 
   d = (struct data *) calloc (1, sizeof (struct data));
-  
+  if (d == NULL)
+    return CL_OUT_OF_HOST_MEMORY;
+
   d->current_kernel = NULL;
   d->current_dlhandle = 0;
   device->data = d;
@@ -417,7 +349,10 @@ pocl_basic_init (cl_device_id device, const char* parameters)
      initialize global_mem_size which it is not yet. Just put 
      a nonzero there for now. */
   device->global_mem_size = 1;
-  pocl_topology_detect_device_info(device);
+  err = pocl_topology_detect_device_info(device);
+  if (err)
+    ret = CL_INVALID_DEVICE;
+
   POCL_INIT_LOCK (d->cq_lock);
   pocl_cpuinfo_detect_device_info(device);
   pocl_set_buffer_image_limits(device);
@@ -429,8 +364,7 @@ pocl_basic_init (cl_device_id device, const char* parameters)
     device->vendor_id =
       magic[0] | magic[1] << 8 | magic[2] << 16 | magic[3] << 24;
 
-  device->vendor_id += device_number;
-  device_number++;
+  device->vendor_id += j;
 
   /* The basic driver represents only one "compute unit" as
      it doesn't exploit multiple hardware threads. Multiple
@@ -446,6 +380,7 @@ pocl_basic_init (cl_device_id device, const char* parameters)
   device->has_64bit_long=0;
   #endif
 
+  return ret;
 }
 
 cl_int
@@ -454,8 +389,7 @@ pocl_basic_alloc_mem_obj (cl_device_id device, cl_mem mem_obj, void* host_ptr)
   void *b = NULL;
   cl_mem_flags flags = mem_obj->flags;
   unsigned i;
-  POCL_MSG_PRINT_INFO("BASIC: alloc_mem_obj, mem %p, dev %d\n", 
-                      mem_obj, device->dev_id);
+  POCL_MSG_PRINT_MEMORY (" mem %p, dev %d\n", mem_obj, device->dev_id);
   /* check if some driver has already allocated memory for this mem_obj 
      in our global address space, and use that*/
   for (i = 0; i < mem_obj->context->num_devices; ++i)
@@ -468,7 +402,9 @@ pocl_basic_alloc_mem_obj (cl_device_id device, cl_mem mem_obj, void* host_ptr)
         {
           mem_obj->device_ptrs[device->dev_id].mem_ptr =
             mem_obj->device_ptrs[i].mem_ptr;
-          POCL_MSG_PRINT_INFO("BASIC: alloc_mem_obj %p dev %d, using already allocated mem\n", mem_obj, device->dev_id);
+          POCL_MSG_PRINT_MEMORY (
+              "mem %p dev %d, using already allocated mem\n", mem_obj,
+              device->dev_id);
           return CL_SUCCESS;
         }
     }
@@ -482,6 +418,7 @@ pocl_basic_alloc_mem_obj (cl_device_id device, cl_mem mem_obj, void* host_ptr)
     }
   else
     {
+      POCL_MSG_PRINT_MEMORY ("!USE_HOST_PTR\n");
       b = pocl_memalign_alloc_global_mem (device, MAX_EXTENDED_ALIGNMENT,
                                           mem_obj->size);
       if (b==NULL)
@@ -491,11 +428,12 @@ pocl_basic_alloc_mem_obj (cl_device_id device, cl_mem mem_obj, void* host_ptr)
     }
 
   /* use this dev mem allocation as host ptr */
-  if (flags & CL_MEM_ALLOC_HOST_PTR && (mem_obj->mem_host_ptr == NULL))
+  if ((flags & CL_MEM_ALLOC_HOST_PTR) && (mem_obj->mem_host_ptr == NULL))
     mem_obj->mem_host_ptr = b;
 
   if (flags & CL_MEM_COPY_HOST_PTR)
     {
+      POCL_MSG_PRINT_MEMORY ("COPY_HOST_PTR\n");
       // mem_host_ptr must be non-NULL
       assert(host_ptr != NULL);
       memcpy (b, host_ptr, mem_obj->size);
@@ -636,7 +574,12 @@ pocl_basic_run
   pc->local_size[0] = cmd->command.run.local_x;
   pc->local_size[1] = cmd->command.run.local_y;
   pc->local_size[2] = cmd->command.run.local_z;
-  
+
+  unsigned rm = pocl_save_rm ();
+  pocl_set_default_rm ();
+  unsigned ftz = pocl_save_ftz ();
+  pocl_set_ftz (kernel->program->flush_denorms);
+
   for (z = 0; z < pc->num_groups[2]; ++z)
     {
       for (y = 0; y < pc->num_groups[1]; ++y)
@@ -652,6 +595,10 @@ pocl_basic_run
             }
         }
     }
+
+  pocl_restore_rm (rm);
+  pocl_restore_ftz (ftz);
+
   for (i = 0; i < kernel->num_args; ++i)
     {
       if (kernel->arg_info[i].is_local)
@@ -860,7 +807,7 @@ void pocl_basic_memfill(void *ptr,
       break;
     case 16:
       {
-      uint64_t * p = (uint64_t*)ptr + offset;
+      uint64_t * p = (uint64_t*)ptr + (offset << 1);
       for (i = 0; i < size; i++)
         for (j = 0; j < 2; j++)
           p[(i<<1) + j] = *((uint64_t*)pattern + j);
@@ -868,7 +815,7 @@ void pocl_basic_memfill(void *ptr,
       break;
     case 32:
       {
-      uint64_t * p = (uint64_t*)ptr + offset;
+      uint64_t * p = (uint64_t*)ptr + (offset << 2);
       for (i = 0; i < size; i++)
         for (j = 0; j < 4; j++)
           p[(i<<2) + j] = *((uint64_t*)pattern + j);
@@ -876,7 +823,7 @@ void pocl_basic_memfill(void *ptr,
       break;
     case 64:
       {
-      uint64_t * p = (uint64_t*)ptr + offset;
+      uint64_t * p = (uint64_t*)ptr + (offset << 3);
       for (i = 0; i < size; i++)
         for (j = 0; j < 8; j++)
           p[(i<<3) + j] = *((uint64_t*)pattern + j);
@@ -884,7 +831,7 @@ void pocl_basic_memfill(void *ptr,
       break;
     case 128:
       {
-      uint64_t * p = (uint64_t*)ptr + offset;
+      uint64_t * p = (uint64_t*)ptr + (offset << 4);
       for (i = 0; i < size; i++)
         for (j = 0; j < 16; j++)
           p[(i<<4) + j] = *((uint64_t*)pattern + j);
@@ -901,9 +848,17 @@ pocl_basic_map_mem (void *data, void *buf_ptr,
                       size_t offset, size_t size,
                       void *host_ptr) 
 {
-  /* All global pointers of the pthread/CPU device are in 
-     the host address space already, and up to date. */
-  if (host_ptr != NULL) return host_ptr;
+  /* If the buffer was allocated with CL_MEM_ALLOC_HOST_PTR |
+   * CL_MEM_COPY_HOST_PTR,
+   * the host_ptr is not the same memory as pocl's device_ptrs[], and we need
+   * to copy pocl's buffer content back to host_ptr. */
+  if ((host_ptr != NULL) && (host_ptr != (buf_ptr + offset)))
+    {
+      POCL_MSG_PRINT_MEMORY ("device: MAP memcpy() "
+                             "buf_ptr %p + offset %zu to host_ptr %p\n",
+                             buf_ptr, offset, host_ptr);
+      memcpy ((char *)host_ptr, (char *)buf_ptr + offset, size);
+    }
   return (char*)buf_ptr + offset;
 }
 
@@ -911,7 +866,18 @@ void* pocl_basic_unmap_mem(void *data, void *host_ptr,
                            void *device_start_ptr,
                            size_t offset, size_t size)
 {
-  return host_ptr;
+  /* If the buffer was allocated with CL_MEM_ALLOC_HOST_PTR |
+   * CL_MEM_COPY_HOST_PTR,
+   * the host_ptr is not the same memory as pocl's device_ptrs[], and we need
+   * to copy host_ptr content back to  pocl's device_ptrs[]. */
+  if ((host_ptr != NULL) && (host_ptr != (device_start_ptr + offset)))
+    {
+      POCL_MSG_PRINT_MEMORY ("device: UNMAP memcpy() "
+                             "host_ptr %p to buf_ptr %p + offset %zu\n",
+                             host_ptr, device_start_ptr, offset);
+      memcpy ((char *)device_start_ptr + offset, (char *)host_ptr, size);
+    }
+  return (char *)host_ptr;
 }
 
 
@@ -951,15 +917,13 @@ static void basic_command_scheduler (struct data *d)
   while ((node = d->ready_list))
     {
       assert (pocl_command_is_ready(node->event));
-      CDL_DELETE (d->ready_list, node);
-
-      
-      pthread_mutex_unlock (&d->cq_lock);
       assert (node->event->status == CL_SUBMITTED);
+      CDL_DELETE (d->ready_list, node);
+      POCL_UNLOCK (d->cq_lock);
       pocl_exec_command(node);
-      pthread_mutex_lock (&d->cq_lock);
+      POCL_LOCK (d->cq_lock);
     }
-    
+
   return;
 }
 
@@ -967,15 +931,16 @@ void
 pocl_basic_submit (_cl_command_node *node, cl_command_queue cq)
 {
   struct data *d = node->device->data;
-  cl_event *event = &(node->event);
   
   node->device->ops->compile_kernel (node, NULL, NULL);
-  POCL_LOCK (d->cq_lock);
-  POCL_UPDATE_EVENT_SUBMITTED(event);
-  pocl_command_push(node, &d->ready_list, &d->command_list);
-  
-  basic_command_scheduler (d);
 
+  POCL_LOCK_OBJ (node->event);
+  node->ready = 1;
+  POCL_LOCK (d->cq_lock);
+  pocl_command_push(node, &d->ready_list, &d->command_list);
+  POCL_UNLOCK_OBJ (node->event);
+
+  basic_command_scheduler (d);
   POCL_UNLOCK (d->cq_lock);
 
   return;
@@ -989,16 +954,7 @@ void pocl_basic_flush (cl_device_id device, cl_command_queue cq)
   basic_command_scheduler (d);
   POCL_UNLOCK (d->cq_lock);
 }
-/*
-static void
-pocl_basic_push_command (_cl_command_node *node)
-{
-  struct data *d = (struct data*)node->device->data;
 
-  pocl_command_push(node, &d->ready_list, &d->command_list);
-
-}
-*/
 void
 pocl_basic_join(cl_device_id device, cl_command_queue cq)
 {
@@ -1012,18 +968,25 @@ pocl_basic_join(cl_device_id device, cl_command_queue cq)
 }
 
 void
-pocl_basic_notify (cl_device_id device, cl_event event)
+pocl_basic_notify (cl_device_id device, cl_event event, cl_event finished)
 {
   struct data *d = (struct data*)device->data;
   _cl_command_node * volatile node = event->command;
-  
-  POCL_LOCK_OBJ (event);
-  if (!(node->ready) && pocl_command_is_ready(node->event))
+
+  if (finished->status < CL_COMPLETE)
     {
-      node->ready = 1;
-      POCL_UNLOCK_OBJ (event);
-      if (node->event->status == CL_SUBMITTED)
+      POCL_UPDATE_EVENT_FAILED (event);
+      return;
+    }
+
+  if (!node->ready)
+    return;
+
+  if (pocl_command_is_ready (event))
+    {
+      if (event->status == CL_QUEUED)
         {
+          POCL_UPDATE_EVENT_SUBMITTED (event);
           POCL_LOCK (d->cq_lock);
           CDL_DELETE (d->command_list, node);
           CDL_PREPEND (d->ready_list, node);
@@ -1032,7 +995,6 @@ pocl_basic_notify (cl_device_id device, cl_event event)
         }
       return;
     }
-  POCL_UNLOCK_OBJ (event);
 }
 
 void
