@@ -22,14 +22,17 @@
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
    THE SOFTWARE.
 */
+
+#define _GNU_SOURCE
+
 #include "common.h"
 #include "pocl_shared.h"
+#include <assert.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
 #include <utlist.h>
-#include <assert.h>
 
 #ifndef _MSC_VER
 #  include <sys/time.h>
@@ -40,14 +43,15 @@
 #endif
 
 #include "config.h"
-#include "pocl_image_util.h"
-#include "pocl_file_util.h"
-#include "pocl_util.h"
-#include "pocl_cache.h"
+#include "config2.h"
 #include "devices.h"
+#include "pocl_cache.h"
+#include "pocl_debug.h"
+#include "pocl_file_util.h"
+#include "pocl_image_util.h"
 #include "pocl_mem_management.h"
 #include "pocl_runtime_config.h"
-#include "pocl_debug.h"
+#include "pocl_util.h"
 
 #ifdef OCS_AVAILABLE
 #include "pocl_llvm.h"
@@ -55,7 +59,8 @@
 
 #include "_kernel_constants.h"
 
-#define COMMAND_LENGTH 2048
+
+#define WORKGROUP_STRING_LENGTH 1024
 
 /**
  * Generate code from the final bitcode using the LLVM
@@ -69,78 +74,159 @@
 
 #ifdef OCS_AVAILABLE
 char*
-llvm_codegen (const char* tmpdir, cl_kernel kernel, cl_device_id device) {
+llvm_codegen (const char* tmpdir, cl_kernel kernel, cl_device_id device,
+              size_t local_x, size_t local_y, size_t local_z)
+{
+  int error = 0;
+  void *write_lock = NULL;
+  void *llvm_module = NULL;
 
-  char command[COMMAND_LENGTH];
-  char bytecode[POCL_FILENAME_LENGTH];
-  char objfile[POCL_FILENAME_LENGTH];
-  /* strlen of / .so 4+1 */
-  int file_name_alloc_size = 
-    min(POCL_FILENAME_LENGTH, strlen(tmpdir) + strlen(kernel->name) + 5);
-  char* module = (char*) malloc(file_name_alloc_size); 
-  /* To avoid corrupted .so files, create a tmp file first
-     and then rename it. */
-  char tmp_module[file_name_alloc_size + 4]; /* .tmp postfix */
-  int error;
+  char tmp_module[POCL_FILENAME_LENGTH];
 
-  error = snprintf(module, POCL_FILENAME_LENGTH,
-                   "%s/%s.so", tmpdir, kernel->name);
+  char objfile_path[POCL_FILENAME_LENGTH];
+  char *objfile = NULL;
+  size_t objfile_size = 0;
 
-  assert (error >= 0);
+  cl_program program = kernel->program;
 
-  error = snprintf(objfile, POCL_FILENAME_LENGTH,
-                   "%s/%s.so.o", tmpdir, kernel->name);
-  assert (error >= 0);
+  int device_i = pocl_cl_device_to_index (program, device);
+  assert (device_i >= 0);
 
-  if (pocl_exists(module))
-    return module;
+  /* $/parallel.bc */
+  char parallel_bc_path[POCL_FILENAME_LENGTH];
+  pocl_cache_work_group_function_path (parallel_bc_path, program, device_i,
+                                       kernel, local_x, local_y, local_z);
 
-  memcpy (tmp_module, module, file_name_alloc_size);
-  strcat (tmp_module, ".tmp");
+  /* $/kernel.so */
+  char final_binary_path[POCL_FILENAME_LENGTH];
+  pocl_cache_final_binary_path (final_binary_path, program, device_i, kernel,
+                                local_x, local_y, local_z);
 
-  void* write_lock = pocl_cache_acquire_writer_lock(kernel->program, device);
-  assert(write_lock);
+  if (pocl_exists (final_binary_path))
+    goto FINISH;
 
-  error = snprintf (bytecode, POCL_FILENAME_LENGTH,
-		    "%s%s", tmpdir, POCL_PARALLEL_BC_FILENAME);
-  assert (error >= 0);
+  /* $/kernel.so.o */
+  assert (strlen (final_binary_path) < (POCL_FILENAME_LENGTH - 3));
+  strcpy (objfile_path, final_binary_path);
+  strcat (objfile_path, ".o");
 
-  error = pocl_llvm_codegen( kernel, device, bytecode, objfile);
-  assert (error == 0);
+  error = pocl_llvm_generate_workgroup_function_nowrite (
+      device, kernel, local_x, local_y, local_z, &llvm_module);
+  if (error)
+    {
+      POCL_MSG_PRINT_GENERAL ("pocl_llvm_generate_workgroup_function() failed"
+                              " for kernel %s\n",
+                              kernel->name);
+      goto FINISH;
+    }
+  assert (llvm_module != NULL);
 
-  /* clang is used as the linker driver in LINK_CMD */
-  error = snprintf (command, COMMAND_LENGTH,
+  /* may happen if another thread is building the same program & wins
+   * the llvm lock. */
+  if (pocl_exists (final_binary_path))
+    goto FINISH;
+
+  error = pocl_llvm_codegen (kernel, device, llvm_module, &objfile,
+                             &objfile_size);
+  if (error)
+    {
+      POCL_MSG_PRINT_GENERAL ("pocl_llvm_codegen() failed"
+                              " for kernel %s\n",
+                              kernel->name);
+      goto FINISH;
+    }
+
+  if (pocl_exists (final_binary_path))
+    goto FINISH;
+
+  /**************************************************************************/
+  write_lock = pocl_cache_acquire_writer_lock (kernel->program, device);
+  assert (write_lock);
+
+  /* write parallel.bc only if we want to leave compiler files*/
+  if (pocl_get_bool_option ("POCL_LEAVE_KERNEL_COMPILER_TEMP_FILES", 0))
+    {
+      POCL_MSG_PRINT_LLVM ("Writing parallel.bc to %s.\n", parallel_bc_path);
+      error = pocl_cache_write_kernel_parallel_bc (
+          llvm_module, program, device_i, kernel, local_x, local_y, local_z);
+    }
+  else
+    {
+      char kernel_parallel_path[POCL_FILENAME_LENGTH];
+      pocl_cache_kernel_cachedir_path (kernel_parallel_path, program, device_i,
+                                       kernel, "", local_x, local_y, local_z);
+      error = pocl_mkdir_p (kernel_parallel_path);
+    }
+  if (error)
+    {
+      POCL_MSG_PRINT_GENERAL ("writing parallel.bc failed"
+                              " for kernel %s\n",
+                              kernel->name);
+      goto FINISH;
+    }
+
+  /* write krenel.so.o always, required for linking step. */
+  POCL_MSG_PRINT_LLVM ("Writing code gen output to %s.\n", objfile_path);
+  error = pocl_write_file (objfile_path, objfile, objfile_size, 0, 0);
+  if (error)
+    {
+      POCL_MSG_PRINT_GENERAL ("writing kernel.so.o failed"
+                              " for kernel %s\n",
+                              kernel->name);
+      goto FINISH;
+    }
+  else
+    {
+      POCL_MSG_PRINT_GENERAL ("written kernel.so.o size %zu\n", objfile_size);
+    }
+
+  /* create a temporary filename */
+  pocl_cache_tempname (tmp_module, ".so", NULL);
+  assert (pocl_exists (tmp_module) > 0);
+
+  /* clang is used as the linker driver for ANDROID, otherwise GNU ld is used
+   */
+  POCL_MSG_PRINT_INFO ("Linking final module\n");
+  char *const args1[]
 #ifndef POCL_ANDROID
-#ifdef OCS_AVAILABLE
-                    CLANGXX " " HOST_CLANG_FLAGS " " HOST_LD_FLAGS " -o %s %s",
+      = { LINK_COMMAND,
+          "-o",
+          tmp_module,
+          objfile_path,
+          HOST_LD_FLAGS_ARRAY,
+          NULL };
 #else
-                    LINK_COMMAND " " HOST_LD_FLAGS " -o %s %s",
+      = { POCL_ANDROID_PREFIX "/bin/ld",
+          "-o",
+          tmp_module,
+          objfile_path,
+          HOST_LD_FLAGS_ARRAY,
+          NULL };
 #endif
-#else
-                    POCL_ANDROID_PREFIX"/bin/ld " HOST_LD_FLAGS " -o %s %s ",
-#endif
-                    tmp_module, objfile);
-  assert (error >= 0);
+  error = pocl_run_command (args1);
+  if (error)
+    goto FINISH;
 
-  POCL_MSG_PRINT_INFO ("executing [%s]\n", command);
-  error = system (command);
-  assert (error == 0);
-
-  error = snprintf (command, COMMAND_LENGTH, "mv %s %s", tmp_module, module);
-  assert (error >= 0);
-  error = system (command);
-  assert (error == 0);
+  error = pocl_rename (tmp_module, final_binary_path);
+  if (error)
+    goto FINISH;
 
   /* Save space in kernel cache */
   if (!pocl_get_bool_option("POCL_LEAVE_KERNEL_COMPILER_TEMP_FILES", 0))
     {
-      pocl_remove(objfile);
-      pocl_remove(bytecode);
+      pocl_remove (objfile_path);
     }
 
-  pocl_cache_release_lock(write_lock);
 
-  return module;
+FINISH:
+  pocl_cache_release_lock (write_lock);
+  pocl_destroy_llvm_module (llvm_module);
+  POCL_MEM_FREE (objfile);
+
+  if (error)
+    return NULL;
+  else
+    return strdup (final_binary_path);
 }
 #endif
 
@@ -154,17 +240,20 @@ fill_dev_image_t (dev_image_t* di, struct pocl_argument* parg,
                   cl_device_id device)
 {
   cl_mem mem = *(cl_mem *)parg->value;
-  di->data = (mem->device_ptrs[device->dev_id].mem_ptr);
   di->width = mem->image_width;
   di->height = mem->image_height;
   di->depth = mem->image_depth;
   di->row_pitch = mem->image_row_pitch;
   di->slice_pitch = mem->image_slice_pitch;
   di->order = mem->image_channel_order;
+  di->image_array_size = mem->image_array_size;
   di->data_type = mem->image_channel_data_type;
   pocl_get_image_information (mem->image_channel_order,
                               mem->image_channel_data_type, &(di->num_channels),
                               &(di->elem_size));
+
+  HANDLE_IMAGE1D_BUFFER (mem);
+  di->data = (mem->device_ptrs[device->dev_id].mem_ptr);
 }
 
 void
@@ -268,6 +357,7 @@ pocl_mem_objs_cleanup (cl_event event)
   event->mem_objs = NULL;
 }
 
+static const size_t zero_origin[] = { 0, 0, 0 };
 /**
  * executes given command.
  */
@@ -276,7 +366,7 @@ pocl_exec_command (_cl_command_node * volatile node)
 {
   unsigned i;
   /* because of POCL_UPDATE_EVENT_ */
-  cl_event *event = &(node->event);
+  cl_event event = node->event;
   switch (node->type)
     {
     case CL_COMMAND_READ_BUFFER:
@@ -286,9 +376,8 @@ pocl_exec_command (_cl_command_node * volatile node)
          node->command.read.host_ptr, 
          node->command.read.device_ptr,
          node->command.read.offset,
-         node->command.read.cb); 
-      POCL_UPDATE_EVENT_COMPLETE(event);
-      POCL_DEBUG_EVENT_TIME(event, "Read Buffer           ");
+         node->command.read.cb);
+      POCL_UPDATE_EVENT_COMPLETE_MSG (event, "Read Buffer           ");
       break;
     case CL_COMMAND_WRITE_BUFFER:
       POCL_UPDATE_EVENT_RUNNING(event);
@@ -298,8 +387,7 @@ pocl_exec_command (_cl_command_node * volatile node)
          node->command.write.device_ptr,
          node->command.write.offset, 
          node->command.write.cb);
-      POCL_UPDATE_EVENT_COMPLETE(event);
-      POCL_DEBUG_EVENT_TIME(event, "Write Buffer          ");
+      POCL_UPDATE_EVENT_COMPLETE_MSG (event, "Write Buffer          ");
       break;
     case CL_COMMAND_COPY_BUFFER:
       POCL_UPDATE_EVENT_RUNNING(event);
@@ -310,37 +398,43 @@ pocl_exec_command (_cl_command_node * volatile node)
                             node->command.copy.src_buffer,
                             node->command.copy.src_offset, 
                             node->command.copy.cb);
-      POCL_UPDATE_EVENT_COMPLETE(event);
-      POCL_DEBUG_EVENT_TIME(event, "Copy Buffer           ");
+      POCL_UPDATE_EVENT_COMPLETE_MSG (event, "Copy Buffer           ");
       break;
     case CL_COMMAND_MIGRATE_MEM_OBJECTS:
       POCL_UPDATE_EVENT_RUNNING(event);
       pocl_migrate_mem_objects (node);
-      POCL_UPDATE_EVENT_COMPLETE(event);
-      POCL_DEBUG_EVENT_TIME(event, "Migrate Buffer        ");
+      POCL_UPDATE_EVENT_COMPLETE_MSG (event, "Migrate Buffer        ");
       break;
     case CL_COMMAND_MAP_IMAGE:
     case CL_COMMAND_MAP_BUFFER: 
       POCL_UPDATE_EVENT_RUNNING(event);
-      pocl_map_mem_cmd (node->device, node->command.map.buffer, 
-                        node->command.map.mapping);
-      POCL_UPDATE_EVENT_COMPLETE(event);
-      POCL_DEBUG_EVENT_TIME(event, "Map Image/Buffer      ");
+      POCL_LOCK_OBJ (node->command.map.buffer);
+      if (node->device->ops->map_mem != NULL)
+        node->device->ops->map_mem (node->device->data,
+                                    (node->command.map.buffer)
+                                        ->device_ptrs[node->device->dev_id]
+                                        .mem_ptr,
+                                    (node->command.map.mapping)->offset,
+                                    (node->command.map.mapping)->size,
+                                    (node->command.map.mapping)->host_ptr);
+      (node->command.map.buffer)->map_count++;
+      POCL_UNLOCK_OBJ (node->command.map.buffer);
+      POCL_UPDATE_EVENT_COMPLETE_MSG (event, "Map Image/Buffer      ");
       break;
     case CL_COMMAND_WRITE_IMAGE:
       POCL_UPDATE_EVENT_RUNNING(event);
-      node->device->ops->write_rect
-        (node->device->data,
+      node->device->ops->write_rect (
+         node->device->data,
          node->command.write_image.host_ptr,
          node->command.write_image.device_ptr,
          node->command.write_image.origin,
-         node->command.write_image.origin,
+         zero_origin,
          node->command.write_image.region,
          node->command.write_image.b_rowpitch,
          node->command.write_image.b_slicepitch,
-         node->command.write_image.b_rowpitch,
-         node->command.write_image.b_slicepitch);
-      POCL_UPDATE_EVENT_COMPLETE(event);
+         node->command.write_image.h_rowpitch,
+         node->command.write_image.h_slicepitch);
+      POCL_UPDATE_EVENT_COMPLETE_MSG (event, "Write Image           ");
       break;
     case CL_COMMAND_WRITE_BUFFER_RECT:
       POCL_UPDATE_EVENT_RUNNING(event);
@@ -355,23 +449,22 @@ pocl_exec_command (_cl_command_node * volatile node)
          node->command.write_image.b_slicepitch,
          node->command.write_image.h_rowpitch,
          node->command.write_image.h_slicepitch);
-      POCL_UPDATE_EVENT_COMPLETE(event);
-      POCL_DEBUG_EVENT_TIME(event, "Write Image           ");
+      POCL_UPDATE_EVENT_COMPLETE_MSG (event, "Write Image           ");
       break;
     case CL_COMMAND_READ_IMAGE:
       POCL_UPDATE_EVENT_RUNNING(event);
-      node->device->ops->read_rect
-        (node->device->data, node->command.read_image.host_ptr,
+      node->device->ops->read_rect (
+         node->device->data,
+         node->command.read_image.host_ptr,
          node->command.read_image.device_ptr,
          node->command.read_image.origin,
-         node->command.read_image.origin,
+         zero_origin,
          node->command.read_image.region,
          node->command.read_image.b_rowpitch,
          node->command.read_image.b_slicepitch,
-         node->command.read_image.b_rowpitch,
-         node->command.read_image.b_slicepitch);
-      POCL_UPDATE_EVENT_COMPLETE(event);
-      POCL_DEBUG_EVENT_TIME(event, "Read Image            ");
+         node->command.read_image.h_rowpitch,
+         node->command.read_image.h_slicepitch);
+      POCL_UPDATE_EVENT_COMPLETE_MSG (event, "Read Image            ");
       break;
     case CL_COMMAND_READ_BUFFER_RECT:
       POCL_UPDATE_EVENT_RUNNING(event);
@@ -385,8 +478,7 @@ pocl_exec_command (_cl_command_node * volatile node)
          node->command.read_image.b_slicepitch,
          node->command.read_image.h_rowpitch,
          node->command.read_image.h_slicepitch);
-      POCL_UPDATE_EVENT_COMPLETE(event);
-      POCL_DEBUG_EVENT_TIME(event, "Read Buffer Rect      ");
+      POCL_UPDATE_EVENT_COMPLETE_MSG (event, "Read Buffer Rect      ");
       break;
     case CL_COMMAND_COPY_BUFFER_RECT:
     case CL_COMMAND_COPY_BUFFER_TO_IMAGE:
@@ -456,52 +548,44 @@ pocl_exec_command (_cl_command_node * volatile node)
              tmp_slicepitch);
           free (tmp);
         }
-      POCL_UPDATE_EVENT_COMPLETE(event);
-      POCL_DEBUG_EVENT_TIME(event, "Copy Buffer Rect      ");
+      POCL_UPDATE_EVENT_COMPLETE_MSG (event, "Copy Buffer Rect      ");
       break;
     case CL_COMMAND_UNMAP_MEM_OBJECT:
       POCL_UPDATE_EVENT_RUNNING(event);
-      if ((node->command.unmap.memobj)->flags & 
-          (CL_MEM_USE_HOST_PTR | CL_MEM_ALLOC_HOST_PTR))
-        {
-          /* TODO: should we ensure the device global region is updated from
-             the host memory? How does the specs define it,
-             can the host_ptr be assumed to point to the host and the
-             device accessible memory or just point there until the
-             kernel(s) get executed or similar? */
-          /* Assume the region is automatically up to date. */
-        } else 
-        {
-          if (node->device->ops->unmap_mem != NULL)        
-            node->device->ops->unmap_mem
-              (node->device->data, 
-               (node->command.unmap.mapping)->host_ptr, 
-               (node->command.unmap.memobj)->device_ptrs[node->device->dev_id].mem_ptr, 
-               (node->command.unmap.mapping)->offset,
-               (node->command.unmap.mapping)->size);
-        }
+      /* TODO: should we ensure the device global region is updated from
+         the host memory? How does the specs define it,
+         can the host_ptr be assumed to point to the host and the
+         device accessible memory or just point there until the
+         kernel(s) get executed or similar? */
+      /* Assume the region is automatically up to date. */
       POCL_LOCK_OBJ (node->command.unmap.memobj);
+      if (node->device->ops->unmap_mem != NULL)
+        node->device->ops->unmap_mem (node->device->data,
+                                      (node->command.unmap.mapping)->host_ptr,
+                                      (node->command.unmap.memobj)
+                                          ->device_ptrs[node->device->dev_id]
+                                          .mem_ptr,
+                                      (node->command.unmap.mapping)->offset,
+                                      (node->command.unmap.mapping)->size);
       DL_DELETE((node->command.unmap.memobj)->mappings, 
                 node->command.unmap.mapping);
       (node->command.unmap.memobj)->map_count--;
+      POCL_MEM_FREE (node->command.unmap.mapping);
       POCL_UNLOCK_OBJ (node->command.unmap.memobj);
-      POCL_UPDATE_EVENT_COMPLETE(event);
-      POCL_DEBUG_EVENT_TIME(event, "Unmap Mem obj         ");
+      POCL_UPDATE_EVENT_COMPLETE_MSG (event, "Unmap Mem obj         ");
       break;
     case CL_COMMAND_NDRANGE_KERNEL:
       POCL_UPDATE_EVENT_RUNNING(event);
-      assert (*event == node->event);
+      assert (event == node->event);
       node->device->ops->run(node->command.run.data, node);
-      POCL_UPDATE_EVENT_COMPLETE(event);
-      POCL_DEBUG_EVENT_TIME(event, "Enqueue NDRange       ");
+      POCL_UPDATE_EVENT_COMPLETE_MSG (event, "Enqueue NDRange       ");
       pocl_ndrange_node_cleanup(node);
       break;
     case CL_COMMAND_NATIVE_KERNEL:
       POCL_UPDATE_EVENT_RUNNING(event);
       node->device->ops->run_native(node->command.native.data, node);
       pocl_native_kernel_cleanup(node);
-      POCL_UPDATE_EVENT_COMPLETE(event);
-      POCL_DEBUG_EVENT_TIME(event, "Native Kernel         ");
+      POCL_UPDATE_EVENT_COMPLETE_MSG (event, "Native Kernel         ");
       break;
     case CL_COMMAND_FILL_IMAGE:
       POCL_UPDATE_EVENT_RUNNING(event);
@@ -514,8 +598,7 @@ pocl_exec_command (_cl_command_node * volatile node)
          node->command.fill_image.slicepitch,
          node->command.fill_image.fill_pixel,
          node->command.fill_image.pixel_size);
-      POCL_UPDATE_EVENT_COMPLETE(event);
-      POCL_DEBUG_EVENT_TIME(event, "Fill Image            ");
+      POCL_UPDATE_EVENT_COMPLETE_MSG (event, "Fill Image            ");
       free(node->command.fill_image.fill_pixel);
       break;
     case CL_COMMAND_FILL_BUFFER:
@@ -526,8 +609,7 @@ pocl_exec_command (_cl_command_node * volatile node)
          node->command.memfill.offset,
          node->command.memfill.pattern,
          node->command.memfill.pattern_size);
-      POCL_UPDATE_EVENT_COMPLETE(event);
-      POCL_DEBUG_EVENT_TIME(event, "Fill Buffer           ");
+      POCL_UPDATE_EVENT_COMPLETE_MSG (event, "Fill Buffer           ");
       pocl_aligned_free(node->command.memfill.pattern);
       break;
     case CL_COMMAND_MARKER:
@@ -550,8 +632,7 @@ pocl_exec_command (_cl_command_node * volatile node)
         for (i=0; i < node->command.svm_free.num_svm_pointers; i++)
           node->device->ops->free_ptr(node->device,
                                       node->command.svm_free.svm_pointers[i]);
-      POCL_UPDATE_EVENT_COMPLETE(event);
-      POCL_DEBUG_EVENT_TIME(event, "SVM Free              ");
+      POCL_UPDATE_EVENT_COMPLETE_MSG (event, "SVM Free              ");
       break;
     case CL_COMMAND_SVM_MAP:
       POCL_UPDATE_EVENT_RUNNING(event);
@@ -561,8 +642,7 @@ pocl_exec_command (_cl_command_node * volatile node)
         node->device->ops->map_mem
           (node->device->data, node->command.svm_map.svm_ptr,
            0, node->command.svm_map.size, NULL);
-      POCL_UPDATE_EVENT_COMPLETE(event);
-      POCL_DEBUG_EVENT_TIME(event, "SVM Map              ");
+      POCL_UPDATE_EVENT_COMPLETE_MSG (event, "SVM Map              ");
       break;
     case CL_COMMAND_SVM_UNMAP:
       POCL_UPDATE_EVENT_RUNNING(event);
@@ -572,8 +652,7 @@ pocl_exec_command (_cl_command_node * volatile node)
         node->device->ops->unmap_mem
           (node->device->data, NULL,
            node->command.svm_unmap.svm_ptr, 0, 0);
-      POCL_UPDATE_EVENT_COMPLETE(event);
-      POCL_DEBUG_EVENT_TIME(event, "SVM Unmap             ");
+      POCL_UPDATE_EVENT_COMPLETE_MSG (event, "SVM Unmap             ");
       break;
     case CL_COMMAND_SVM_MEMCPY:
       POCL_UPDATE_EVENT_RUNNING(event);
@@ -581,8 +660,7 @@ pocl_exec_command (_cl_command_node * volatile node)
                               node->command.svm_memcpy.src, 0,
                               node->command.svm_memcpy.dst, 0,
                               node->command.svm_memcpy.size);
-      POCL_UPDATE_EVENT_COMPLETE(event);
-      POCL_DEBUG_EVENT_TIME(event, "SVM Memcpy            ");
+      POCL_UPDATE_EVENT_COMPLETE_MSG (event, "SVM Memcpy            ");
       break;
     case CL_COMMAND_SVM_MEMFILL:
       POCL_UPDATE_EVENT_RUNNING(event);
@@ -591,8 +669,7 @@ pocl_exec_command (_cl_command_node * volatile node)
                                  node->command.memfill.size, 0,
                                  node->command.memfill.pattern,
                                  node->command.memfill.pattern_size);
-      POCL_UPDATE_EVENT_COMPLETE(event);
-      POCL_DEBUG_EVENT_TIME(event, "SVM MemFill           ");
+      POCL_UPDATE_EVENT_COMPLETE_MSG (event, "SVM MemFill           ");
       pocl_aligned_free(node->command.memfill.pattern);
       break;
     default:
@@ -607,9 +684,13 @@ pocl_broadcast (cl_event brc_event)
 {
   event_node *target;
   event_node *tmp;
+
   while ((target = brc_event->notify_list))
     {
-      POCL_LOCK_OBJ (target->event);
+      if (brc_event->command_type == CL_COMMAND_USER)
+        POCL_LOCK_OBJ (target->event);
+      else
+        pocl_lock_events_inorder (brc_event, target->event);
       /* remove event from wait list */
       LL_FOREACH (target->event->wait_list, tmp)
         {
@@ -620,17 +701,20 @@ pocl_broadcast (cl_event brc_event)
               break;
             }
         }
-      if (target->event->status == CL_SUBMITTED)
-        {
+        if (brc_event->command_type == CL_COMMAND_USER)
           POCL_UNLOCK_OBJ (target->event);
-          target->event->command->device->ops->notify 
-            (target->event->command->device, target->event);
-        }
-      else 
-        POCL_UNLOCK_OBJ (target->event);
-      
-      LL_DELETE (brc_event->notify_list, target);
-      pocl_mem_manager_free_event_node (target);
+
+        if ((target->event->status == CL_SUBMITTED)
+            || (target->event->status == CL_QUEUED))
+          {
+            target->event->command->device->ops->notify (
+                target->event->command->device, target->event, brc_event);
+          }
+
+        LL_DELETE (brc_event->notify_list, target);
+        if (brc_event->command_type != CL_COMMAND_USER)
+          pocl_unlock_events_inorder (brc_event, target->event);
+        pocl_mem_manager_free_event_node (target);
     }
 }
 
@@ -641,13 +725,13 @@ pocl_broadcast (cl_event brc_event)
 void
 fill_dev_sampler_t (dev_sampler_t *ds, struct pocl_argument *parg)
 {
-  cl_sampler_t sampler = *(cl_sampler_t *)parg->value;
+  cl_sampler sampler = *(cl_sampler *)parg->value;
 
-  *ds = 0;
-  *ds |= sampler.normalized_coords == CL_TRUE ? CLK_NORMALIZED_COORDS_TRUE :
-      CLK_NORMALIZED_COORDS_FALSE;
+  *ds = (sampler->normalized_coords == CL_TRUE) ? CLK_NORMALIZED_COORDS_TRUE
+                                                : CLK_NORMALIZED_COORDS_FALSE;
 
-  switch (sampler.addressing_mode) {
+  switch (sampler->addressing_mode)
+    {
     case CL_ADDRESS_NONE:
       *ds |= CLK_ADDRESS_NONE; break;
     case CL_ADDRESS_CLAMP_TO_EDGE:
@@ -660,7 +744,8 @@ fill_dev_sampler_t (dev_sampler_t *ds, struct pocl_argument *parg)
       *ds |= CLK_ADDRESS_MIRRORED_REPEAT; break;
   }
 
-  switch (sampler.filter_mode) {
+  switch (sampler->filter_mode)
+    {
     case CL_FILTER_NEAREST:
       *ds |= CLK_FILTER_NEAREST; break;
     case CL_FILTER_LINEAR :
@@ -719,7 +804,7 @@ static int handle_count = 0;
 void
 pocl_check_dlhandle_cache (_cl_command_node *cmd)
 {
-  char workgroup_string[256];
+  char workgroup_string[WORKGROUP_STRING_LENGTH];
   pocl_dlhandle_cache_item *ci = NULL;
 
   POCL_LOCK (pocl_dlhandle_cache_lock);
@@ -770,7 +855,10 @@ pocl_check_dlhandle_cache (_cl_command_node *cmd)
       POCL_LOCK (pocl_llvm_codegen_lock);
       module_fn = (char *)llvm_codegen (cmd->command.run.tmp_dir,
                                         cmd->command.run.kernel,
-                                        cmd->device);
+                                        cmd->device,
+                                        cmd->command.run.local_x,
+                                        cmd->command.run.local_y,
+                                        cmd->command.run.local_z);
       POCL_UNLOCK (pocl_llvm_codegen_lock);
       POCL_MSG_PRINT_INFO("Using static WG size binary: %s\n", module_fn);
 #else
@@ -813,7 +901,8 @@ pocl_check_dlhandle_cache (_cl_command_node *cmd)
     }
   free(module_fn);
 
-  snprintf (workgroup_string, 256, "_pocl_launcher_%s_workgroup", 
+  snprintf (workgroup_string, WORKGROUP_STRING_LENGTH,
+            "_pocl_launcher_%s_workgroup",
             cmd->command.run.kernel->name);
 
   POCL_LOCK (pocl_dlhandle_lock);
@@ -829,44 +918,26 @@ pocl_check_dlhandle_cache (_cl_command_node *cmd)
   POCL_UNLOCK (pocl_dlhandle_cache_lock);
 }
 
-/*
-static void
-pocl_free_dlhandle (_cl_command_node *cmd)
-{
-  pocl_dlhandle_cache_item *ci = NULL;
-  POCL_LOCK (pocl_dlhandle_cache_lock);
-  DL_FOREACH (pocl_dlhandle_cache, ci)
-    {
-      if (strcmp (ci->tmp_dir, cmd->command.run.tmp_dir) == 0 &&
-          strcmp (ci->function_name, 
-                  cmd->command.run.kernel->name) == 0)
-        {
-          if ((--ci->ref_count))
-            break;
-          --handle_count;
-          DL_DELETE (pocl_dlhandle_cache, ci);
-          POCL_UNLOCK (pocl_dlhandle_cache_lock);
-          free (ci->tmp_dir);
-          free (ci->function_name);
-          POCL_LOCK (pocl_llvm_codegen_lock);
-          assert(!lt_dlclose (ci->dlhandle));
-          POCL_UNLOCK (pocl_llvm_codegen_lock);
-          free (ci);
-          return;
-        }
-    }
-  POCL_UNLOCK (pocl_dlhandle_cache_lock);
-}
-*/
 
 #define MIN_MAX_MEM_ALLOC_SIZE (128*1024*1024)
 
 /* accounting object for the main memory */
 static pocl_global_mem_t system_memory;
 
+static size_t
+next_larger_pow2 (size_t in)
+{
+  size_t out = 1;
+  while (in > out)
+    out <<= 1;
+  return out;
+}
+
 void
 pocl_setup_device_for_system_memory(cl_device_id device)
 {
+  int limit_memory_gb = pocl_get_int_option ("POCL_MEMORY_LIMIT", 0);
+
   /* set up system memory limits, if required */
   if (system_memory.total_alloc_limit == 0)
   {
@@ -877,7 +948,7 @@ pocl_setup_device_for_system_memory(cl_device_id device)
        */
       size_t alloc_limit = device->global_mem_size;
       if ((alloc_limit >> 20) > (7 << 10))
-        system_memory.total_alloc_limit = alloc_limit - (size_t)(1 << 31);
+        system_memory.total_alloc_limit = alloc_limit - (size_t)(1UL << 31);
       else
         {
           size_t temp = (alloc_limit >> 2);
@@ -889,6 +960,18 @@ pocl_setup_device_for_system_memory(cl_device_id device)
   }
 
   device->global_mem_size = system_memory.total_alloc_limit;
+
+  if (limit_memory_gb > 0)
+    {
+      size_t limited_memory = (size_t)limit_memory_gb << 30;
+      if (device->global_mem_size > limited_memory)
+        device->global_mem_size = limited_memory;
+      else
+        POCL_MSG_WARN ("requested POCL_MEMORY_LIMIT %i GBs is larger than "
+                       "physical memory size (%zu) GBs, ignoring\n",
+                       limit_memory_gb, (device->global_mem_size >> 30));
+    }
+
   if (device->global_mem_size < MIN_MAX_MEM_ALLOC_SIZE)
     POCL_ABORT("Not enough memory to run on this device.\n");
 
@@ -909,7 +992,9 @@ pocl_setup_device_for_system_memory(cl_device_id device)
     alloc_limit = MIN_MAX_MEM_ALLOC_SIZE;
 
   if (alloc_limit > device->global_mem_size)
-    alloc_limit = device->global_mem_size;
+    alloc_limit = next_larger_pow2 (device->global_mem_size / 4);
+  if (alloc_limit > (device->global_mem_size / 2))
+    alloc_limit >>= 1;
 
   if (alloc_limit < MIN_MAX_MEM_ALLOC_SIZE)
     alloc_limit = MIN_MAX_MEM_ALLOC_SIZE;
@@ -929,9 +1014,13 @@ pocl_set_buffer_image_limits(cl_device_id device)
 {
   pocl_setup_device_for_system_memory(device);
   /* these aren't set up in pocl_setup_device_for_system_memory,
-   * because some devices (HSA) set them up themselves */
-  device->local_mem_size = device->max_constant_buffer_size =
-      device->max_mem_alloc_size;
+   * because some devices (HSA) set them up themselves
+   *
+   * it's max mem alloc / 4 because some programs (conformance test)
+   * try to allocate max size constant objects and run out of memory
+   * while trying to fill them. */
+  device->local_mem_size = device->max_constant_buffer_size
+      = next_larger_pow2 (device->global_mem_size / 256);
 
   /* We don't have hardware limitations on the buffer-backed image sizes,
    * so we set the maximum size in terms of the maximum amount of pixels
@@ -981,11 +1070,13 @@ pocl_memalign_alloc_global_mem(cl_device_id device, size_t align, size_t size)
   if (!ptr)
     return NULL;
 
+  POCL_LOCK_OBJ (mem);
   mem->currently_allocated += size;
   if (mem->max_ever_allocated < mem->currently_allocated)
     mem->max_ever_allocated = mem->currently_allocated;
-
   assert(mem->currently_allocated <= mem->total_alloc_limit);
+  POCL_UNLOCK_OBJ (mem);
+
   return ptr;
 }
 
@@ -994,8 +1085,10 @@ pocl_free_global_mem(cl_device_id device, void* ptr, size_t size)
 {
   pocl_global_mem_t *mem = device->global_memory;
 
+  POCL_LOCK_OBJ (mem);
   assert(mem->currently_allocated >= size);
   mem->currently_allocated -= size;
+  POCL_UNLOCK_OBJ (mem);
 
   POCL_MEM_FREE(ptr);
 }
@@ -1003,7 +1096,7 @@ pocl_free_global_mem(cl_device_id device, void* ptr, size_t size)
 void
 pocl_print_system_memory_stats()
 {
-  POCL_MSG_PRINT("MEM STATS:\n", "",
+  POCL_MSG_PRINT_F (MEMORY, INFO, "",
   "____ Total available system memory  : %10zu KB\n"
   " ____ Currently used system memory   : %10zu KB\n"
   " ____ Max used system memory         : %10zu KB\n",
