@@ -21,16 +21,14 @@
    THE SOFTWARE.
 */
 
+#include <errno.h>
 #include <stddef.h>
 #include <stdio.h>
 #include <string.h>
 #include <unistd.h>
 
 #include "config.h"
-
-#ifdef POCL_BUILT_WITH_CMAKE
 #include "pocl_build_timestamp.h"
-#endif
 
 #ifdef OCS_AVAILABLE
 #include "kernellib_hash.h"
@@ -233,32 +231,86 @@ void* pocl_cache_acquire_reader_lock(cl_program program,
 
 /******************************************************************************/
 
-void pocl_cache_mk_temp_name(char* path) {
-    assert(cache_topdir_initialized);
+static void
+pocl_cache_mk_temp_name (char *path_template, unsigned suffix_len, int *ret_fd)
+{
+  assert (cache_topdir_initialized);
 #if defined(_MSC_VER) || defined(__MINGW32__)
     char* tmp = _tempnam(cache_topdir, "pocl_");
     assert(tmp);
-    int bytes_written = snprintf(path, POCL_FILENAME_LENGTH, "%s", tmp);
+    int bytes_written
+        = snprintf (path_template, POCL_FILENAME_LENGTH, "%s", tmp);
     free(tmp);
     assert(bytes_written > 0 && bytes_written < POCL_FILENAME_LENGTH);
 #else
-    int bytes_written = snprintf(path, POCL_FILENAME_LENGTH,
-             "%s/temp_XXXXXX.cl", cache_topdir);
-    assert(bytes_written > 0 && bytes_written < POCL_FILENAME_LENGTH);
     /* using mkstemp() instead of tmpnam() has no real benefit
      * here, as we have to pass the filename to llvm,
      * but tmpnam() generates an annoying warning... */
-    int fd = mkstemps(path, 3);
-    assert(fd >= 0);
-    close(fd);
+    int fd;
+
+    if (suffix_len)
+      fd = mkstemps (path_template, suffix_len);
+    else
+      fd = mkstemp (path_template);
+
+    if (fd < 0)
+      {
+        char buf[512];
+        strerror_r (errno, buf, 512);
+        POCL_ABORT ("mkstemp failed: %s\n", buf);
+      }
+
+    if (ret_fd)
+      *ret_fd = fd;
+    else
+      close (fd);
+
+    return;
 #endif
 }
 
-int pocl_cache_write_program_source(char *program_cl_path,
-                                    cl_program program) {
-    pocl_cache_mk_temp_name(program_cl_path);
-    return pocl_write_file(program_cl_path, program->source,
-                           strlen(program->source), 0, 0);
+int
+pocl_cache_create_tempdir (char *path)
+{
+  assert (cache_topdir_initialized);
+#if defined(_MSC_VER) || defined(__MINGW32__)
+  char *tmp = _tempnam (cache_topdir, "pocl_");
+  assert (tmp);
+  int bytes_written = snprintf (path, POCL_FILENAME_LENGTH, "%s", tmp);
+  free (tmp);
+  assert (bytes_written > 0 && bytes_written < POCL_FILENAME_LENGTH);
+  return 0;
+#else
+  int bytes_written = snprintf (path, POCL_FILENAME_LENGTH,
+                                "%s/tempdir_XXXXXX", cache_topdir);
+  assert (bytes_written > 0 && bytes_written < POCL_FILENAME_LENGTH);
+  /* TODO mkdtemp() might not be portable */
+  return (mkdtemp (path) == NULL);
+#endif
+}
+
+void
+pocl_cache_tempname (char *path_template, const char *suffix, int *fd)
+{
+  assert (cache_topdir_initialized);
+  assert (path_template);
+  strcpy (path_template, cache_topdir);
+  size_t suffixlen = (suffix ? strlen (suffix) : 0);
+  size_t max = POCL_FILENAME_LENGTH - 16 - suffixlen;
+  assert (strlen (path_template) < max);
+  strcat (path_template, "/tempfile_XXXXXX");
+  if (suffix)
+    strcat (path_template, suffix);
+
+  pocl_cache_mk_temp_name (path_template, suffixlen, fd);
+}
+
+int
+pocl_cache_write_program_source (char *program_cl_path, cl_program program)
+{
+  pocl_cache_tempname (program_cl_path, ".cl", NULL);
+  return pocl_write_file (program_cl_path, program->source,
+                          strlen (program->source), 0, 0);
 }
 
 /******************************************************************************/
@@ -333,6 +385,9 @@ int pocl_cache_append_to_buildlog(cl_program  program,
                                   unsigned    device_i,
                                   const char *content,
                                   size_t      size) {
+    if (!buildhash_is_valid (program, device_i))
+      return -1;
+
     char buildlog_path[POCL_FILENAME_LENGTH];
     program_device_dir(buildlog_path, program,
                        device_i, POCL_BUILDLOG_FILENAME);
@@ -392,14 +447,24 @@ build_program_compute_hash(cl_program program,
       pocl_SHA1_Update(&hash_ctx,
 		       (uint8_t*) program->pocl_binaries[device_i],
 		       program->pocl_binary_sizes[device_i]);
-
-    } else {
-      /* Program was created with clCreateProgramWithBinary() with an LLVM IR binary */
-      assert(program->binaries[device_i]);
-      pocl_SHA1_Update(&hash_ctx,
-		       (uint8_t*) program->binaries[device_i],
-		       program->binary_sizes[device_i]);
-    }
+      }
+    else if (program->binary_sizes[device_i] > 0)
+      {
+        /* Program was created with clCreateProgramWithBinary() with an LLVM IR
+         * binary */
+        assert (program->binaries[device_i]);
+        pocl_SHA1_Update (&hash_ctx, (uint8_t *)program->binaries[device_i],
+                          program->binary_sizes[device_i]);
+      }
+    else
+      {
+        /* Program is linked from binaries, has no source or binary */
+        // assert(program->binary_type == CL_PROGRAM_BIN)
+        assert (preprocessed_source);
+        assert (source_len > 0);
+        pocl_SHA1_Update (&hash_ctx, (uint8_t *)preprocessed_source,
+                          source_len);
+      }
 
     if (program->compiler_options)
         pocl_SHA1_Update(&hash_ctx, (uint8_t*) program->compiler_options,
@@ -481,16 +546,19 @@ char* pocl_get_process_name()
 
 /******************************************************************************/
 
-void pocl_cache_init_topdir() {
+int
+pocl_cache_init_topdir ()
+{
 
-    if (cache_topdir_initialized)
-        return;
+  if (cache_topdir_initialized)
+    return 0;
 
-    const char *tmp_path = pocl_get_string_option("POCL_CACHE_DIR", NULL);
-    int needed;
+  const char *tmp_path = pocl_get_string_option ("POCL_CACHE_DIR", NULL);
+  int needed;
 
-    if (tmp_path && (pocl_exists(tmp_path))) {
-        needed = snprintf(cache_topdir, POCL_FILENAME_LENGTH, "%s", tmp_path);
+  if (tmp_path)
+    {
+      needed = snprintf (cache_topdir, POCL_FILENAME_LENGTH, "%s", tmp_path);
     } else     {
 #ifdef POCL_ANDROID
         char* process_name = pocl_get_process_name();
@@ -530,15 +598,33 @@ void pocl_cache_init_topdir() {
 #endif
     }
 
-    if (needed >= POCL_FILENAME_LENGTH) {
-        POCL_ABORT("pocl: cache path longer than maximum filename length\n");
+  if (needed >= POCL_FILENAME_LENGTH)
+    {
+      POCL_MSG_ERR ("pocl: cache path longer than maximum filename length\n");
+      return 1;
     }
 
     assert(strlen(cache_topdir) > 0);
-    if (pocl_mkdir_p(cache_topdir))
-        POCL_ABORT("Could not create topdir %s for cache\n", cache_topdir);
-    cache_topdir_initialized = 1;
 
+    if (pocl_mkdir_p(cache_topdir))
+      {
+        POCL_MSG_ERR (
+            "Could not create top directory (%s) for cache. \n\nNote: "
+            "if you have proper rights to create that directory, and still "
+            "get the error, then most likely pocl and the program you're "
+            "trying to run are linked to different versions of libstdc++ "
+            "library. \nThis is not a bug in pocl and there's nothing we "
+            "can do to fix it - you need both pocl and your program to be"
+            " compiled for your system. This is known to happen with "
+            "Luxmark benchmark binaries dowloaded from website; Luxmark "
+            "installed from your linux distribution's packages should "
+            "work.\n",
+            cache_topdir);
+        return 1;
+      }
+
+    cache_topdir_initialized = 1;
+    return 0;
 }
 
 /* Create the new program cachedir, invalidating the old program
@@ -624,8 +710,8 @@ void pocl_cache_cleanup_cachedir(cl_program program) {
             void* lock = acquire_program_lock(program, i, "_read", 0);
             if (!lock)
               {
-                POCL_MSG_PRINT(" *** WARNING *** ", "",
-                "Could not get an exclusive lock to remove program cachedir");
+                POCL_MSG_WARN ("Could not get an exclusive lock "
+                               "to remove program cachedir\n");
                 continue;
               }
             char cachedir[POCL_FILENAME_LENGTH];
