@@ -26,14 +26,18 @@
 #include <stdlib.h>
 #include <string.h>
 
-#include <sys/types.h>
-#include <sys/stat.h>
 #include <time.h>
 
 #ifndef _MSC_VER
-#  include <dirent.h>
-#  include <unistd.h>
-#  include <utime.h>
+#include <dirent.h>
+#include <sys/resource.h>
+#include <sys/stat.h>
+#include <sys/time.h>
+#include <sys/types.h>
+#include <sys/wait.h>
+#include <unistd.h>
+#include <unistd.h>
+#include <utime.h>
 #else
 #  include "vccompat.hpp"
 #endif
@@ -46,6 +50,11 @@
 #include "devices.h"
 #include "pocl_runtime_config.h"
 
+/* required for setting SSE/AVX flush denorms to zero flag */
+#if defined(__x86_64__) && defined(__GNUC__)
+#include <x86intrin.h>
+#endif
+
 struct list_item;
 
 typedef struct list_item
@@ -54,6 +63,110 @@ typedef struct list_item
   struct list_item *next;
 } list_item;
 
+
+
+void
+pocl_restore_ftz (unsigned ftz)
+{
+#if defined(__x86_64__) && defined(__GNUC__)
+
+#ifdef _MM_FLUSH_ZERO_ON
+  if (ftz & _MM_FLUSH_ZERO_ON)
+    _MM_SET_FLUSH_ZERO_MODE (_MM_FLUSH_ZERO_ON);
+  else
+    _MM_SET_FLUSH_ZERO_MODE (_MM_FLUSH_ZERO_OFF);
+#endif
+#ifdef _MM_DENORMALS_ZERO_ON
+  if (ftz & _MM_DENORMALS_ZERO_ON)
+    _MM_SET_DENORMALS_ZERO_MODE (_MM_DENORMALS_ZERO_ON);
+  else
+    _MM_SET_DENORMALS_ZERO_MODE (_MM_DENORMALS_ZERO_OFF);
+#endif
+
+#endif
+}
+
+unsigned
+pocl_save_ftz ()
+{
+#if defined(__x86_64__) && defined(__GNUC__)
+
+  unsigned s = 0;
+#ifdef _MM_FLUSH_ZERO_ON
+  if (_MM_GET_FLUSH_ZERO_MODE ())
+    s |= _MM_FLUSH_ZERO_ON;
+  else
+    s &= (~_MM_FLUSH_ZERO_ON);
+#endif
+#ifdef _MM_DENORMALS_ZERO_ON
+  if (_MM_GET_DENORMALS_ZERO_MODE ())
+    s |= _MM_DENORMALS_ZERO_ON;
+  else
+    s &= (~_MM_DENORMALS_ZERO_ON);
+#endif
+  return s;
+
+#else
+  return 0;
+#endif
+}
+
+void
+pocl_set_ftz (unsigned ftz)
+{
+#if defined(__x86_64__) && defined(__GNUC__)
+  if (ftz)
+    {
+#ifdef _MM_FLUSH_ZERO_ON
+      _MM_SET_FLUSH_ZERO_MODE (_MM_FLUSH_ZERO_ON);
+#endif
+
+#ifdef _MM_DENORMALS_ZERO_ON
+      _MM_SET_DENORMALS_ZERO_MODE (_MM_DENORMALS_ZERO_ON);
+#endif
+    }
+  else
+    {
+#ifdef _MM_FLUSH_ZERO_OFF
+      _MM_SET_FLUSH_ZERO_MODE (_MM_FLUSH_ZERO_OFF);
+#endif
+
+#ifdef _MM_DENORMALS_ZERO_OFF
+      _MM_SET_DENORMALS_ZERO_MODE (_MM_DENORMALS_ZERO_OFF);
+#endif
+    }
+#endif
+}
+
+
+void
+pocl_set_default_rm ()
+{
+#if defined(__x86_64__) && defined(__GNUC__) && defined(_MM_ROUND_NEAREST)
+  unsigned rm = _MM_GET_ROUNDING_MODE ();
+  if (rm != _MM_ROUND_NEAREST)
+    _MM_SET_ROUNDING_MODE (_MM_ROUND_NEAREST);
+#endif
+}
+
+unsigned
+pocl_save_rm ()
+{
+#if defined(__x86_64__) && defined(__GNUC__) && defined(_MM_ROUND_NEAREST)
+  return _MM_GET_ROUNDING_MODE ();
+#else
+  return 0;
+#endif
+}
+
+void
+pocl_restore_rm (unsigned rm)
+{
+#if defined(__x86_64__) && defined(__GNUC__) && defined(_MM_ROUND_NEAREST)
+  _MM_SET_ROUNDING_MODE (rm);
+#endif
+}
+
 uint32_t
 byteswap_uint32_t (uint32_t word, char should_swap)
 {
@@ -179,6 +292,35 @@ pocl_aligned_free (void *ptr)
 }
 #endif
 
+void
+pocl_lock_events_inorder (cl_event ev1, cl_event ev2)
+{
+  if (ev1->id < ev2->id)
+    {
+      POCL_LOCK_OBJ (ev1);
+      POCL_LOCK_OBJ (ev2);
+    }
+  else
+    {
+      POCL_LOCK_OBJ (ev2);
+      POCL_LOCK_OBJ (ev1);
+    }
+}
+
+void
+pocl_unlock_events_inorder (cl_event ev1, cl_event ev2)
+{
+  if (ev1->id < ev2->id)
+    {
+      POCL_UNLOCK_OBJ (ev1);
+      POCL_UNLOCK_OBJ (ev2);
+    }
+  else
+    {
+      POCL_UNLOCK_OBJ (ev2);
+      POCL_UNLOCK_OBJ (ev1);
+    }
+}
 
 cl_int pocl_create_event (cl_event *event, cl_command_queue command_queue, 
                           cl_command_type command_type, int num_buffers,
@@ -186,12 +328,14 @@ cl_int pocl_create_event (cl_event *event, cl_command_queue command_queue,
 {
   static unsigned int event_id_counter = 0;
 
+  POCL_MSG_PRINT_EVENTS ("creating event\n");
+
   if (context == NULL || !(context->valid))
     return CL_INVALID_CONTEXT;
   if (event != NULL)
     {
       *event = pocl_mem_manager_new_event ();
-      if (event == NULL)
+      if (*event == NULL)
         return CL_OUT_OF_HOST_MEMORY;
 
       (*event)->context = context;
@@ -203,23 +347,14 @@ cl_int pocl_create_event (cl_event *event, cl_command_queue command_queue,
         POname(clRetainCommandQueue) (command_queue);
 
       (*event)->command_type = command_type;
-      (*event)->command =  NULL;
-      (*event)->callback_list = NULL;
       (*event)->id = event_id_counter++;
-      (*event)->notify_list = NULL;
-      (*event)->data = NULL;
       (*event)->num_buffers = num_buffers;
       if (num_buffers > 0)
         {
           (*event)->mem_objs = malloc (num_buffers * sizeof(cl_mem));
           memcpy ((*event)->mem_objs, buffers, num_buffers * sizeof(cl_mem));
         }
-      else
-        (*event)->mem_objs = NULL;
       (*event)->status = CL_QUEUED;
-      (*event)->implicit_event = 0;
-      (*event)->next = NULL;
-      (*event)->prev = NULL;
 
       /* user events do not have cq */
       if (!command_queue)
@@ -235,6 +370,10 @@ pocl_create_event_sync(cl_event waiting_event,
 {
   event_node * volatile notify_target = NULL;
   event_node * volatile wait_list_item = NULL;
+
+  if (notifier_event == NULL)
+    return CL_SUCCESS;
+
   assert(notifier_event->pocl_refcount != 0);
   POCL_MSG_PRINT_INFO("create event sync: waiting %d, notifier %d\n", waiting_event->id, notifier_event->id);
   if (waiting_event == notifier_event)
@@ -243,19 +382,17 @@ pocl_create_event_sync(cl_event waiting_event,
              notifier_event->id);
       assert(waiting_event != notifier_event);
     }
+
+  pocl_lock_events_inorder (waiting_event, notifier_event);
+
   LL_FOREACH (waiting_event->wait_list, wait_list_item)
     {
       if (wait_list_item->event == notifier_event)
-        return CL_SUCCESS;
-    }
-  POCL_LOCK_OBJ (waiting_event);
-
-  if (notifier_event == NULL || notifier_event->status == CL_COMPLETE)
-    {
-      POCL_UNLOCK_OBJ (waiting_event);
-      return CL_SUCCESS;
+        goto FINISH;
     }
 
+  if (notifier_event->status == CL_COMPLETE)
+    goto FINISH;
   notify_target = pocl_mem_manager_new_event_node();
   wait_list_item = pocl_mem_manager_new_event_node();
   if (!notify_target || !wait_list_item)
@@ -265,8 +402,9 @@ pocl_create_event_sync(cl_event waiting_event,
   wait_list_item->event = notifier_event;
   LL_PREPEND (notifier_event->notify_list, notify_target);
   LL_PREPEND (waiting_event->wait_list, wait_list_item);
-  POCL_UNLOCK_OBJ (waiting_event);
 
+FINISH:
+  pocl_unlock_events_inorder (waiting_event, notifier_event);
   return CL_SUCCESS;
 }
 
@@ -315,6 +453,7 @@ cl_int pocl_create_command (_cl_command_node **cmd,
      one reference for the host and one for the runtime/driver */
   if (event_p)
     {
+      POCL_MSG_PRINT_EVENTS ("event pointer provided\n");
       *event_p = *event;
       (*event)->implicit_event = 0;
       (*event)->pocl_refcount = 2;
@@ -325,11 +464,8 @@ cl_int pocl_create_command (_cl_command_node **cmd,
       (*event)->pocl_refcount = 1;
     }
 
-  (*cmd)->next = NULL;
-  (*cmd)->prev = NULL;
   (*cmd)->device = command_queue->device;
   (*cmd)->event->command = (*cmd);
-  (*cmd)->ready = 0;
 
   /* in case of in-order queue, synchronize to previously enqueued command
      if available */
@@ -338,11 +474,9 @@ cl_int pocl_create_command (_cl_command_node **cmd,
       POCL_LOCK_OBJ (command_queue);
       if (command_queue->last_event.event)
         {
-          POCL_LOCK_OBJ (command_queue->last_event.event);
           pocl_create_event_sync ((*cmd)->event,
                                   command_queue->last_event.event,
                                   NULL);
-          POCL_UNLOCK_OBJ (command_queue->last_event.event);
         }
       POCL_UNLOCK_OBJ (command_queue);
     }
@@ -350,12 +484,10 @@ cl_int pocl_create_command (_cl_command_node **cmd,
   for (i = 0; i < num_events; ++i)
     {
       cl_event wle = wait_list[i];
-      POCL_LOCK_OBJ (wle);
       pocl_create_event_sync ((*cmd)->event, wle, NULL);
-      POCL_UNLOCK_OBJ (wle);
     }
-  POCL_MSG_PRINT_INFO("Created command struct (event %d, type %X)\n", 
-                      (*cmd)->event->id, command_type);
+  POCL_MSG_PRINT_EVENTS ("Created command struct (event %d, type %X)\n",
+                         (*cmd)->event->id, command_type);
   return CL_SUCCESS;
 }
 
@@ -377,9 +509,7 @@ void pocl_command_enqueue (cl_command_queue command_queue,
     {
       DL_FOREACH (command_queue->events, event)
         {
-          POCL_LOCK_OBJ(event);
           pocl_create_event_sync (node->event, event, NULL);
-          POCL_UNLOCK_OBJ (event);
         }
     }
   if (node->type == CL_COMMAND_BARRIER)
@@ -388,9 +518,7 @@ void pocl_command_enqueue (cl_command_queue command_queue,
     {
       if (command_queue->barrier)
         {
-          POCL_LOCK_OBJ(command_queue->barrier);
           pocl_create_event_sync (node->event, command_queue->barrier, NULL);
-          POCL_UNLOCK_OBJ (command_queue->barrier);
         }
     }
   DL_APPEND (command_queue->events, node->event);
@@ -398,7 +526,9 @@ void pocl_command_enqueue (cl_command_queue command_queue,
   command_queue->last_event.event_id = node->event->id;
   POCL_UNLOCK_OBJ (command_queue);
 
-  POCL_UPDATE_EVENT_QUEUED (&node->event);
+  POCL_LOCK_OBJ (node->event);
+  POCL_UPDATE_EVENT_QUEUED (node->event);
+  POCL_UNLOCK_OBJ (node->event);
 
   command_queue->device->ops->submit(node, command_queue);
 #ifdef POCL_DEBUG_BUILD
@@ -423,16 +553,15 @@ pocl_command_push (_cl_command_node *node,
       CDL_PREPEND ((*pending_list), node);
       return;
     }
-  POCL_LOCK_OBJ (node->event);
   if (pocl_command_is_ready(node->event))
     {
+      POCL_UPDATE_EVENT_SUBMITTED (node->event);
       CDL_PREPEND ((*ready_list), node);
     }
   else
     {
       CDL_PREPEND ((*pending_list), node);
     }
-  POCL_UNLOCK_OBJ (node->event);
 }
 
 int pocl_update_command_queue (cl_event event)
@@ -455,6 +584,42 @@ int pocl_update_command_queue (cl_event event)
   return cq_ready;
 }
 
+void
+pocl_cl_mem_inherit_flags (cl_mem mem, cl_mem from_buffer, cl_mem_flags flags)
+{
+  if ((flags & CL_MEM_READ_WRITE) | (flags & CL_MEM_READ_ONLY)
+      | (flags & CL_MEM_WRITE_ONLY))
+    {
+      mem->flags = (flags & CL_MEM_READ_WRITE) | (flags & CL_MEM_READ_ONLY)
+                   | (flags & CL_MEM_WRITE_ONLY);
+    }
+  else
+    {
+      mem->flags = (from_buffer->flags & CL_MEM_READ_WRITE)
+                   | (from_buffer->flags & CL_MEM_READ_ONLY)
+                   | (from_buffer->flags & CL_MEM_WRITE_ONLY);
+    }
+
+  if ((flags & CL_MEM_HOST_NO_ACCESS) | (flags & CL_MEM_HOST_READ_ONLY)
+      | (flags & CL_MEM_HOST_WRITE_ONLY))
+    {
+      mem->flags = mem->flags | ((flags & CL_MEM_HOST_NO_ACCESS)
+                                 | (flags & CL_MEM_HOST_READ_ONLY)
+                                 | (flags & CL_MEM_HOST_WRITE_ONLY));
+    }
+  else
+    {
+      mem->flags
+          = mem->flags | ((from_buffer->flags & CL_MEM_HOST_NO_ACCESS)
+                          | (from_buffer->flags & CL_MEM_HOST_READ_ONLY)
+                          | (from_buffer->flags & CL_MEM_HOST_WRITE_ONLY));
+    }
+
+  mem->flags = mem->flags | (from_buffer->flags & CL_MEM_USE_HOST_PTR)
+               | (from_buffer->flags & CL_MEM_ALLOC_HOST_PTR)
+               | (from_buffer->flags & CL_MEM_COPY_HOST_PTR);
+}
+
 cl_int pocl_update_mem_obj_sync (cl_command_queue cq, _cl_command_node *cmd, 
                                  cl_mem mem, char operation)
 {
@@ -476,12 +641,14 @@ cl_int pocl_update_mem_obj_sync (cl_command_queue cq, _cl_command_node *cmd,
 }
 
 int pocl_buffer_boundcheck(cl_mem buffer, size_t offset, size_t size) {
-  POCL_RETURN_ERROR_ON((offset > buffer->size), CL_INVALID_VALUE,
-            "offset(%zu) > buffer->size(%zu)", offset, buffer->size);
-  POCL_RETURN_ERROR_ON((size > buffer->size), CL_INVALID_VALUE,
-            "size(%zu) > buffer->size(%zu)", size, buffer->size);
-  POCL_RETURN_ERROR_ON((offset + size > buffer->size), CL_INVALID_VALUE,
-            "offset + size (%zu) > buffer->size(%zu)", (offset+size), buffer->size);
+  POCL_RETURN_ERROR_ON ((offset > buffer->size), CL_INVALID_VALUE,
+                        "offset(%zu) > buffer->size(%zu)\n", offset,
+                        buffer->size);
+  POCL_RETURN_ERROR_ON ((size > buffer->size), CL_INVALID_VALUE,
+                        "size(%zu) > buffer->size(%zu)\n", size, buffer->size);
+  POCL_RETURN_ERROR_ON ((offset + size > buffer->size), CL_INVALID_VALUE,
+                        "offset + size (%zu) > buffer->size(%zu)\n",
+                        (offset + size), buffer->size);
   return CL_SUCCESS;
 }
 
@@ -575,16 +742,16 @@ int pocl_buffers_overlap(cl_mem src_buffer,
   /* sub buffers overlap check  */
   if (src_buffer->parent && dst_buffer->parent &&
         (src_buffer->parent == dst_buffer->parent)) {
-      src_offset = (char*)src_buffer->mem_host_ptr - (char*)src_buffer->parent->mem_host_ptr +
-        src_offset;
-      dst_offset = (char*)dst_buffer->mem_host_ptr - (char*)dst_buffer->parent->mem_host_ptr +
-        dst_offset;
+      src_offset = src_buffer->origin + src_offset;
+      dst_offset = dst_buffer->origin + dst_offset;
 
-    POCL_RETURN_ERROR_ON(((src_offset <= dst_offset) && (dst_offset <=
-      (src_offset + size - 1))), CL_MEM_COPY_OVERLAP, "dst_offset lies inside \
+      POCL_RETURN_ERROR_ON (((src_offset <= dst_offset)
+                             && (dst_offset <= (src_offset + size - 1))),
+                            CL_MEM_COPY_OVERLAP, "dst_offset lies inside \
       the src region and src_buffer + dst_buffer are subbuffers of the same buffer");
-    POCL_RETURN_ERROR_ON(((dst_offset <= src_offset) && (src_offset <=
-      (dst_offset + size - 1))), CL_MEM_COPY_OVERLAP, "src_offset lies inside \
+      POCL_RETURN_ERROR_ON (((dst_offset <= src_offset)
+                             && (src_offset <= (dst_offset + size - 1))),
+                            CL_MEM_COPY_OVERLAP, "src_offset lies inside \
       the dst region and src_buffer + dst_buffer are subbuffers of the same buffer");
 
   }
@@ -682,6 +849,15 @@ check_copy_overlap(const size_t src_offset[3],
   return overlap;
 }
 
+/* For a subdevice parameter, return the actual device it belongs to. */
+cl_device_id
+pocl_real_dev (const cl_device_id dev)
+{
+  cl_device_id ret = dev;
+  while (ret->parent_device)
+    ret = ret->parent_device;
+  return ret;
+}
 
 /* Make a list of unique devices. If any device is a subdevice,
  * replace with parent, then remove duplicate parents. */
@@ -694,7 +870,7 @@ cl_device_id * pocl_unique_device_list(const cl_device_id * in, cl_uint num, cl_
 
   unsigned i;
   for (i=0; i < num; ++i)
-    out[i] = (in[i] ? POCL_REAL_DEV(in[i]) : NULL);
+    out[i] = (in[i] ? pocl_real_dev (in[i]) : NULL);
 
   i=1;
   unsigned device_i=0;
@@ -741,6 +917,35 @@ void pocl_setup_context(cl_context context)
         }
 }
 
+int
+pocl_check_event_wait_list (cl_command_queue command_queue,
+                            cl_uint num_events_in_wait_list,
+                            const cl_event *event_wait_list)
+{
+  POCL_RETURN_ERROR_COND (
+      (event_wait_list == NULL && num_events_in_wait_list > 0),
+      CL_INVALID_EVENT_WAIT_LIST);
+
+  POCL_RETURN_ERROR_COND (
+      (event_wait_list != NULL && num_events_in_wait_list == 0),
+      CL_INVALID_EVENT_WAIT_LIST);
+
+  if (event_wait_list)
+    {
+      unsigned i;
+      for (i = 0; i < num_events_in_wait_list; i++)
+        {
+          POCL_RETURN_ERROR_COND ((event_wait_list[i] == NULL),
+                                  CL_INVALID_EVENT_WAIT_LIST);
+          POCL_RETURN_ERROR_COND (
+              (event_wait_list[i]->context != command_queue->context),
+              CL_INVALID_CONTEXT);
+        }
+    }
+
+  return CL_SUCCESS;
+}
+
 const char*
 pocl_status_to_str (int status)
 {
@@ -824,3 +1029,130 @@ pocl_command_to_str (cl_command_type cmd)
 
   return "unknown";
 }
+
+/*
+ * This replaces a simple system(), because:
+ *
+ * 1) system() was causing issues (gpu lockups) with HSA when
+ * compiling code (via compile_parallel_bc_to_brig)
+ * with OpenCL 2.0 atomics (like CalcPie from AMD SDK).
+ * The reason of lockups is unknown (yet).
+ *
+ * 2) system() uses fork() which copies page table maps, and runs
+ * out of AS when pocl has already allocated huge buffers in memory.
+ * this happened in llvm_codegen()
+ *
+ * vfork() does not copy pagetables.
+ */
+int
+pocl_run_command (char *const *args)
+{
+  POCL_MSG_PRINT_INFO ("Launching: %s\n", args[0]);
+#ifdef HAVE_VFORK
+  pid_t p = vfork ();
+#elif defined(HAVE_FORK)
+  pid_t p = fork ();
+#else
+#error Must have fork() or vfork() system calls for HSA
+#endif
+  if (p == 0)
+    {
+      return execv (args[0], args);
+    }
+  else
+    {
+      if (p < 0)
+        return EXIT_FAILURE;
+      int status;
+      if (waitpid (p, &status, 0) < 0)
+        POCL_ABORT ("pocl: waitpid() failed.\n");
+      if (WIFEXITED (status))
+        return WEXITSTATUS (status);
+      else if (WIFSIGNALED (status))
+        return WTERMSIG (status);
+      else
+        return EXIT_FAILURE;
+    }
+}
+
+/*
+ * float 2 half / half 2 float
+ */
+
+static int const shift = 13;
+static int const shiftSign = 16;
+
+static int32_t const infN = 0x7F800000;  /* flt32 infinity */
+static int32_t const maxN = 0x477FE000;  /* max flt16 normal as a flt32 */
+static int32_t const minN = 0x38800000;  /* min flt16 normal as a flt32 */
+static int32_t const signN = 0x80000000; /* flt32 sign bit */
+
+/* static int32_t const infC = infN >> shift;
+ * static int32_t const infC = 0x3FC00;
+ * static int32_t const nanN = (infC + 1) << shift; // minimum flt16 nan as a flt32
+ */
+static int32_t const nanN = 0x7f802000;
+/* static int32_t const maxC = maxN >> shift; */
+static int32_t const maxC = 0x23bff;
+/* static int32_t const minC = minN >> shift;
+ * static int32_t const minC = 0x1c400;
+ * static int32_t const signC = signN >> shiftSign; // flt16 sign bit
+ */
+static int32_t const signC = 0x40000; /* flt16 sign bit */
+
+static int32_t const mulN = 0x52000000; /* (1 << 23) / minN */
+static int32_t const mulC = 0x33800000; /* minN / (1 << (23 - shift)) */
+
+static int32_t const subC = 0x003FF; /* max flt32 subnormal down shifted */
+static int32_t const norC = 0x00400; /* min flt32 normal down shifted */
+
+/* static int32_t const maxD = infC - maxC - 1; */
+static int32_t const maxD = 0x1c000;
+/* static int32_t const minD = minC - subC - 1; */
+static int32_t const minD = 0x1c000;
+
+typedef union
+{
+  float f;
+  int32_t si;
+  uint32_t ui;
+} H2F_Bits;
+
+float
+half_to_float (uint16_t value)
+{
+  H2F_Bits v;
+  v.ui = value;
+  int32_t sign = v.si & signC;
+  v.si ^= sign;
+  sign <<= shiftSign;
+  v.si ^= ((v.si + minD) ^ v.si) & -(v.si > subC);
+  v.si ^= ((v.si + maxD) ^ v.si) & -(v.si > maxC);
+  H2F_Bits s;
+  s.si = mulC;
+  s.f *= v.si;
+  int32_t mask = -(norC > v.si);
+  v.si <<= shift;
+  v.si ^= (s.si ^ v.si) & mask;
+  v.si |= sign;
+  return v.f;
+}
+
+uint16_t
+float_to_half (float value)
+{
+  H2F_Bits v, s;
+  v.f = value;
+  uint32_t sign = v.si & signN;
+  v.si ^= sign;
+  sign >>= shiftSign;
+  s.si = mulN;
+  s.si = s.f * v.f;
+  v.si ^= (s.si ^ v.si) & -(minN > v.si);
+  v.si ^= (infN ^ v.si) & -((infN > v.si) & (v.si > maxN));
+  v.si ^= (nanN ^ v.si) & -((nanN > v.si) & (v.si > infN));
+  v.ui >>= shift;
+  v.si ^= ((v.si - maxD) ^ v.si) & -(v.si > maxC);
+  v.si ^= ((v.si - minD) ^ v.si) & -(v.si > subC);
+  return v.ui | sign;
+}
